
module mpi_helpers
  use mpi
  implicit none
  private

  ! Helper functions for easily creating Stucts, Vectors/Arrays, 
  ! Arrays of Structs, Stucts with Vectors, etc.
  ! * Define_MPI_Struct()    - for structs (you probably want to use this) 
  ! * Define_MPI_Vector()    - for arrays / vectors
  ! * Defien_MPI_Empty_Set() - usefull when there is no data to send
  public :: Define_MPI_Struct
  public :: Define_MPI_Vector
  public :: Define_MPI_Empty_Set

contains

  ! Defines a Struct consisting of the elements defined in the array "types"
  ! Example:
  !   newtype = Define_MPI_Struct([MPI_INTEGER,MPI_LOGICAL,some_other_mpi_type,...])
  function Define_MPI_Struct(types) result(newtype)
    implicit none
    integer :: newtype
    integer, dimension(:), intent(in) :: types

    integer, dimension(size(types)) :: blocklengths
    integer(kind=MPI_ADDRESS_KIND), dimension(size(types)) :: displacements
    integer(kind=MPI_ADDRESS_KIND) :: lb,extent
    integer :: ierr,ie2

    integer :: i

    displacements(1) = 0
    blocklengths(1)  = 1
    do i=1,size(types)-1
      call MPI_TYPE_GET_EXTENT(types(i), lb, extent, ierr)
        if (ierr .ne. MPI_SUCCESS) call MPI_ABORT(MPI_COMM_WORLD, ierr, ie2)
      displacements(i+1) = displacements(i) + extent
      blocklengths(i+1) = 1
    enddo
    call MPI_TYPE_CREATE_STRUCT(size(types), blocklengths, displacements, &
                               &types, newtype, ierr)
      if (ierr .ne. MPI_SUCCESS) call MPI_ABORT(MPI_COMM_WORLD, ierr, ie2)
    call MPI_TYPE_COMMIT(newtype, ierr)
      if (ierr .ne. MPI_SUCCESS) call MPI_ABORT(MPI_COMM_WORLD, ierr, ie2)
  end function Define_MPI_Struct
  ! Defines a Vector/Array consisting of N elements of the type "typ"
  ! Example:
  !   newtype1 = Define_MPI_Vector(MPI_INTEGER, 10)
  !   newtype2 = Define_MPI_Vector(some_other_mpi_type, N)
  function Define_MPI_Vector(typ,N) result(newtype)
    implicit none
    integer :: newtype
    integer, intent(in) :: typ
    integer, intent(in) :: N

    integer :: ierr,ie2
    
    call MPI_TYPE_CONTIGUOUS(N, typ, newtype, ierr)
      if (ierr .ne. MPI_SUCCESS) call MPI_ABORT(MPI_COMM_WORLD, ierr, ie2)
    call MPI_TYPE_COMMIT(newtype, ierr)
      if (ierr .ne. MPI_SUCCESS) call MPI_ABORT(MPI_COMM_WORLD, ierr, ie2)
  end function Define_MPI_Vector
  ! Defines a mpi-type which consists of no data. (for example an empty struct).
  ! You can use this, when you don't want to send data in a certain communication-step
  ! Example:
  !   newtype = Define_MPI_Empty_Set()
  function Define_MPI_Empty_Set() result(newtype)
    implicit none
    integer :: newtype
    newtype = Define_MPI_Vector(MPI_INTEGER,0)
  end function Define_MPI_Empty_Set

end module mpi_helpers
