! TODO: More documentation
module mpi_queue
  use mpi
  use mpi_task_list
  use my_data_types
  implicit none

  private ! default is private

  ! public subroutines and properties
  public :: Queue_Init
  public :: Queue_Define_Types
  public :: Queue_Start
  public :: Queue_Finalize
  public :: Queue_Add_Task
  public :: Queue_Set_Global
  ! other getter-functions
  public :: MPI_get_numprocs
  public :: MPI_get_rank

  ! callback-functions
  public :: Queue_Tasks_Initialize
  public :: Queue_Task_Compute
  public :: Queue_Task_Recieve
  public :: Queue_Tasks_Analysis
  abstract interface
    subroutine Queue_Tasks_Initialize_Interface()
      implicit none
    end subroutine Queue_Tasks_Initialize_Interface
    subroutine Queue_Task_Compute_Interface(dat,global, res)
      use my_data_types
      implicit none
      type(task_data), intent(in) :: dat
      type(global_data), intent(in) :: global
      type(task_result), intent(out) :: res
    end subroutine Queue_Task_Compute_Interface
    subroutine Queue_Task_Recieve_Interface(res,task_id)
      use my_data_types
      implicit none
      type(task_result), intent(in) :: res
      integer, intent(in) :: task_id
    end subroutine Queue_Task_Recieve_Interface
    subroutine Queue_Tasks_Analysis_Interface()
      implicit none
    end subroutine Queue_Tasks_Analysis_Interface
  end interface
  procedure (Queue_Tasks_Initialize_Interface), pointer :: Queue_Tasks_Initialize
  procedure (Queue_Task_Compute_Interface),     pointer :: Queue_Task_Compute
  procedure (Queue_Task_Recieve_Interface),     pointer :: Queue_Task_Recieve
  procedure (Queue_Tasks_Analysis_Interface),   pointer :: Queue_Tasks_Analysis

  ! private variables
  ! MPI - variables
  integer :: numprocs
  integer :: rank
  integer :: ierr
  integer :: queue_comm
  ! MPI - data-types
  integer :: mpi_global_data
  integer :: mpi_task_data
  integer :: mpi_task_result
  ! internal variables
  logical :: initialized = .false.
  type(list) :: tasks
  type(global_data) :: global

contains

  subroutine Error(error_code,error_str)
    implicit none
    integer, intent(in) :: error_code
    character(*), intent(in), optional :: error_str
    character(MPI_MAX_ERROR_STRING) :: error_string
    integer :: error_string_length
    if (present(error_str)) then
      ! write custom error-message
      write(*,*) trim(error_str)
    else
      ! Returns a string for a given error code
      call MPI_ERROR_STRING(error_code, error_string, error_string_length, ierr)
      write(*,*) error_string(1:error_string_length)
    endif
    ! Terminates MPI execution environment
    call MPI_ABORT(queue_comm, error_code, ierr)
  end subroutine Error

  subroutine Queue_Init()
    implicit none
    integer :: group
    if (initialized) then
      call Error(-1,'Error: Queue has already been initialized!')
      return
    endif
    ! Initializes the MPI execution environment
    call MPI_INIT(ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
    ! Determines the rank of the calling process in the communicator
    call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
    ! Returns the size of the group associated with a communicator
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numprocs, ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
    ! Create own MPI-communicator
    ! Extract the group handle
    call MPI_COMM_GROUP(MPI_COMM_WORLD, group, ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
    ! Create a new communicator based on the same group as queue_comm
    call MPI_COMM_CREATE(MPI_COMM_WORLD, group, queue_comm, ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
    ! create task list
    call list_create(tasks)
    ! nullify callback-pointers
    nullify(Queue_Tasks_Initialize)
    nullify(Queue_Task_Compute)
    nullify(Queue_Task_Recieve)
    initialized = .true.
  end subroutine Queue_Init

  subroutine Queue_Define_Types()
    implicit none
    if (.not. initialized) then
      call Error(-1,'Error: Queue has not been initialized!')
      return
    endif
    mpi_global_data = Define_MPI_Global_Data()
    mpi_task_data   = Define_MPI_Task_Data()
    mpi_task_result = Define_MPI_Task_Result()
  end subroutine Queue_Define_Types

  subroutine Queue_Start()
    implicit none
    ! check initalization status
    if (.not. initialized) then
      call Error(-1,'Error: Queue has not been initialized!')
      return
    endif
    ! check callback-procedure-pointers
    if (.not.(associated(Queue_Tasks_Initialize) .and. &
              associated(Queue_Task_Compute) .and. &
              associated(Queue_Task_Recieve))) then
      call Error(-1,'Error: At least one non-optional callback-functions has not been defined!')
      return
    endif
    ! Callback: Queue_Tasks_Initialize. This lets the Master add initial Tasks
    if (rank .eq. 0) then
      call Queue_Tasks_Initialize()
    endif
    ! Distribute global data to every process
    call MPI_BCAST(global, 1, mpi_global_data, 0, queue_comm, ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
    ! Begin of actual communication and calculation
    if (numprocs > 1) then 
      ! default mpi-based master-slave algorithm
      if (rank .eq. 0) then
        call Master()
      else
        call Slave()
      endif
    else ! (numprocs = 1)
      ! degenerate algorithm for single process sequential execution
      call Degenerated()
    endif
    ! Callback: Queue_Tasks_Analysis.
    if (rank .eq.0) then
      if (associated(Queue_Tasks_Analysis)) then
        call Queue_Tasks_Analysis()
      endif
    endif
    ! End of communication and calculation
  contains
    ! this handles most of the internal communication. It sends single tasks to the
    ! slaves, waits for results and sends new tasks until the task-list is empty
    subroutine Master()
      implicit none
      integer :: i
      integer :: n
      integer :: tag
      integer :: req_dummy ! dummy variable for unused mpi-request-IDs
      integer :: req_index ! Index of handle for operation that completed
      integer, dimension(1:MPI_STATUS_SIZE) :: mpi_status
      integer, dimension(:), allocatable :: requests
      type(task_data) :: task
      type(task_data), dimension(:), allocatable :: task_buffer
      type(task_result), dimension(:), allocatable :: results

      n = list_num_elements(tasks)
      ! allocate send and recieve buffers. 
      ! This allows us to use non-blocking sends and recieves. :)
      allocate(task_buffer(min(numprocs-1,n)))
      allocate(results    (min(numprocs-1,n)))
      allocate(requests   (min(numprocs-1,n)))
      ! send one task to each process and post non blocking recieves
      do i=1,min(numprocs-1,n)
        call list_dequeue(tasks, task, tag)
        task_buffer(i) = task
        ! register the "recieving" of the results
        call MPI_IRECV(results(i), 1, mpi_task_result, i, tag,&
                       queue_comm, requests(i), ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
        ! send and do not wait until the send-buffer is reusable (as we 
        ! have our own "task_buffer", we can use non-blocking sends)
        call MPI_ISEND(task_buffer(i), 1, mpi_task_data, i, tag, &
                       queue_comm, req_dummy, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
      enddo
      ! distribute remaining datasets
      do i=min(numprocs-1,n)+1,n
        call MPI_WAITANY(min(numprocs-1,n), requests, req_index, mpi_status, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
        ! recieve - callback
        call Queue_Task_Recieve(results(mpi_status(MPI_SOURCE)), mpi_status(MPI_TAG))
        ! now send new vector to process number mpi_status(MPI_SOURCE)
        call list_dequeue(tasks, task, tag)
        task_buffer(mpi_status(MPI_SOURCE)) = task
        call MPI_IRECV(results(mpi_status(MPI_SOURCE)), 1, mpi_task_result, &
                       mpi_status(MPI_SOURCE), tag, queue_comm, &
                       requests(mpi_status(MPI_SOURCE)), ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
        call MPI_ISEND(task_buffer(mpi_status(MPI_SOURCE)), 1, mpi_task_data, &
                       mpi_status(MPI_SOURCE), tag, queue_comm, &
                       req_dummy, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
      enddo
      ! let all processes come to an end
      do i=1,min(numprocs-1,n)
        call MPI_WAITANY(min(numprocs-1,n), requests, req_index, mpi_status, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
        ! recieve - callback
        call Queue_Task_Recieve(results(mpi_status(MPI_SOURCE)), mpi_status(MPI_TAG))
        ! send termination signal
        call MPI_ISEND(task, 1, mpi_task_data, mpi_status(MPI_SOURCE), 0,&
                       queue_comm, req_dummy, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
      enddo
      do i=min(numprocs-1,n)+1,numprocs-1
        ! send termination signal
        call MPI_ISEND(task, 1, mpi_task_data, i, 0, queue_comm, req_dummy, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
      enddo
      ! Don't forget to deallocate
      deallocate(task_buffer)
      deallocate(results)
      deallocate(requests)
      ! TODO: dynamic addition of tasks? do loops -> while loop with list_is_empty(tasks)?
    end subroutine Master
    ! the slave just waits for new data, calculates the results and returns them. The 
    ! process is halted only after a stop-signal is recieved from the master
    subroutine Slave()
      implicit none
      integer, dimension(1:MPI_STATUS_SIZE) :: mpi_status
      logical :: not_finished
      type(task_data) :: dat
      type(task_result) :: res
      not_finished = .true.
      ! recieve data and calculate
      do while (not_finished)
        ! first recieve ...
        call MPI_RECV(dat, 1, mpi_task_data, 0, MPI_ANY_TAG, queue_comm, &
                      mpi_status, ierr)
          if (ierr .ne. MPI_SUCCESS) call Error(ierr)
        not_finished = (mpi_status(MPI_TAG) .ne. 0)
        if (not_finished) then
          ! ... then calculate ... (compute-callback)
          call Queue_Task_Compute(dat,global, res)
          ! ... and finally send result back
          call MPI_SEND(res, 1, mpi_task_result, 0, mpi_status(MPI_TAG), &
                        queue_comm, ierr)
            if (ierr .ne. MPI_SUCCESS) call Error(ierr)
        endif
      enddo
    end subroutine Slave
    subroutine Degenerated()
      implicit none
      type(task_data)   :: dat
      type(task_result) :: res
      integer :: i,n,tag
      n = list_num_elements(tasks)
      do i=1,n
        call list_dequeue(tasks, dat, tag)
        call Queue_Task_Compute(dat, global, res)
        call Queue_Task_Recieve(res, tag)
      enddo
    end subroutine Degenerated
  end subroutine Queue_Start

  subroutine Queue_Finalize()
    implicit none
    ! destroy task list
    call list_delete(tasks)
    ! Terminates MPI execution environment
    call MPI_FINALIZE(ierr)
      if (ierr .ne. MPI_SUCCESS) call Error(ierr)
  end subroutine Queue_Finalize

  subroutine Queue_Add_Task(task,task_id)
    implicit none
    type(task_data), intent(in) :: task
    integer, intent(in) :: task_id
    ! add task to task list
    if (task_id <= 0) then
      call Error(-1,'Error: "tag" has to be a positive integer and cannot be zero')
    else
      call list_push(tasks,task,task_id)
    endif
  end subroutine Queue_Add_Task
  
  subroutine Queue_Set_Global(g)
    implicit none
    type(global_data), intent(in) :: g

    global = g
  end subroutine Queue_Set_Global

  ! ---------- other getter functions -----------
  integer function MPI_get_numprocs()
    implicit none
    MPI_get_numprocs = numprocs
  end function MPI_get_numprocs
  integer function MPI_get_rank()
    implicit none
    MPI_get_rank = rank
  end function MPI_get_rank


end module mpi_queue
